from rasa.core.events import (
    UserUtteranceReverted,
    UserUttered,
    ActionExecuted,
    Event,
    SlotSet,
)
from rasa.core.actions.action import Action, ActionUtterTemplate, create_bot_utterance

from typing import Any, List, Text, Dict, Optional
import logging
import copy

from requests.auth import HTTPBasicAuth

logging.basicConfig(level="WARN")
logger = logging.getLogger()


class ActionBotfrontDisambiguation(Action):
    def name(self):
        return "action_botfront_disambiguation"

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
    ) -> List[Event]:
        message = None
        for event in list(tracker.events)[::-1]:
            logger.debug(event)
            if isinstance(event, SlotSet) and event.key == "disambiguation_message":
                message = event.value
                break
        if not message: return []
        template = await nlg.generate(
            message["template"],
            tracker,
            output_channel.name(),
        )
        return [
            create_bot_utterance({
                "text": template.get("text", ""),
                "quick_replies": message.get("quick_replies", []) + template.get("quick_replies", []),
            })
        ]


class ActionBotfrontDisambiguationFollowup(Action):
    def name(self) -> Text:
        return "action_botfront_disambiguation_followup"

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
    ) -> List[Event]:

        revert_events = [
            UserUtteranceReverted(),
            UserUtteranceReverted(),
            ActionExecuted(action_name="action_listen"),
        ]

        last_user_event = None
        for event in list(tracker.events)[::-1]:
            if isinstance(event, UserUttered):
                last_user_event = copy.deepcopy(event)
                last_user_event.parse_data["intent"]["confidence"] = 1.0
                break
        if last_user_event:
            revert_events += [last_user_event]

        return revert_events


class ActionBotfrontFallback(ActionUtterTemplate):
    def name(self) -> Text:
        return "action_botfront_fallback"

    def __init__(self):
        super(ActionBotfrontFallback, self).__init__("utter_fallback", silent_fail=True)

    async def run(self, output_channel, nlg, tracker, domain):

        evts = await super(ActionBotfrontFallback, self).run(
            output_channel, nlg, tracker, domain
        )
        if (
            len(tracker.events) >= 4
            and isinstance(tracker.events[-4], ActionExecuted)
            and tracker.events[-4].action_name == "action_botfront_disambiguation"
        ):
            return evts + [UserUtteranceReverted(), UserUtteranceReverted()]
        else:
            return evts + [UserUtteranceReverted()]




from rasa_addons.core.channels.facebook_messenger import FBMessengerInput
from rasa_addons.core.channels.rest import BotfrontRestInput
from rasa_addons.core.channels.rest_plus import BotfrontRestPlusInput
from rasa_addons.core.channels.webchat import WebchatInput
from rasa_addons.core.channels.webchat_plus import WebchatPlusInput


import time
import logging
import requests
from rasa.nlu.extractors.duckling_http_extractor import DucklingHTTPExtractor, convert_duckling_format_to_rasa
from rasa.utils.common import raise_warning
from rasa.nlu.training_data import Message

from typing import Any, List, Optional, Text, Dict
from rasa.constants import DOCS_URL_COMPONENTS
from rasa.nlu.constants import ENTITIES

logger = logging.getLogger(__name__)


class DucklingHTTPExtractorWithTimezone(DucklingHTTPExtractor):

    def _duckling_parse(self, text: Text, reference_time: int, timezone) -> List[Dict[Text, Any]]:
        """Sends the request to the duckling server and parses the result."""

        try:
            payload = self._payload(text, reference_time)
            payload["tz"] = timezone # mod
            headers = {
                "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8"
            }
            response = requests.post(
                self._url() + "/parse",
                data=payload,
                headers=headers,
                timeout=self.component_config.get("timeout"),
            )
            if response.status_code == 200:
                return response.json()
            else:
                logger.error(
                    "Failed to get a proper response from remote "
                    "duckling. Status Code: {}. Response: {}"
                    "".format(response.status_code, response.text)
                )
                return []
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            logger.error(
                "Failed to connect to duckling http server. Make sure "
                "the duckling server is running/healthy/not stale and the proper host "
                "and port are set in the configuration. More "
                "information on how to run the server can be found on "
                "github: "
                "https://github.com/facebook/duckling#quickstart "
                "Error: {}".format(e)
            )
            return []

    @staticmethod
    def _timezone_from_config_or_request(component_config, timezone):
        if timezone is not None:
            return timezone
        return component_config.get("timezone")

    @staticmethod
    def _reference_time_from_message_or_request(message, reference_time):
        if message.time is not None:
            try:
                return int(message.time) * 1000
            except ValueError as e:
                logging.warning(
                    "Could not parse timestamp {}. Instead "
                    "current UTC time will be passed to "
                    "duckling. Error: {}".format(message.time, e)
                )
        # fallbacks to current time, multiplied by 1000 because duckling
        # requires the reftime in milliseconds
        elif reference_time is not None:
            try:
                return int(reference_time)
            except ValueError as e:
                logging.warning(
                    "Could not parse timestamp {}. Instead "
                    "current UTC time will be passed to "
                    "duckling. Error: {}".format(reference_time, e)
                )
        return int(time.time()) * 1000

    def process(self, message: Message, **kwargs: Any) -> None:

        if self._url() is not None:
            # mod >
            params = kwargs
            timezone = self._timezone_from_config_or_request(
                self.component_config, params.get("timezone", None)
            )
            reference_time = self._reference_time_from_message_or_request(
                message, params.get("reference_time", None)
            )
            matches = self._duckling_parse(message.text, reference_time, timezone)
            # </ mod
            all_extracted = convert_duckling_format_to_rasa(matches)
            dimensions = self.component_config["dimensions"]
            extracted = DucklingHTTPExtractor.filter_irrelevant_entities(
                all_extracted, dimensions
            )
        else:
            extracted = []
            raise_warning(
                "Duckling HTTP component in pipeline, but no "
                "`url` configuration in the config "
                "file nor is `RASA_DUCKLING_HTTP_URL` "
                "set as an environment variable. No entities will be extracted!",
                docs=DOCS_URL_COMPONENTS + "#ducklinghttpextractor",
            )

        extracted = self.add_extractor_name(extracted)
        message.set(
            ENTITIES, message.get(ENTITIES, []) + extracted, add_to_output=True,
        )


from rasa.core.events import Event
from rasa.core.actions.action import Action, create_bot_utterance

from typing import Any, List, Text, Dict, Optional
import logging
import copy

from requests.auth import HTTPBasicAuth

logging.basicConfig(level="WARN")
logger = logging.getLogger()


class ActionBotfrontMapping(Action):
    def name(self):
        return "action_botfront_mapping"

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
    ) -> List[Event]:
        """Append 'utter_' to intent name and generates from that template"""

        events = []
        response_name = "utter_" + tracker.latest_message.intent["name"]
        response = await nlg.generate(
            response_name,
            tracker,
            output_channel.name(),
        )
        events += [create_bot_utterance(response)]

        return events



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import warnings

from builtins import str
from typing import Any, Dict, Optional, Text

from rasa.nlu.extractors.entity_synonyms import EntitySynonymMapper


class EntitySynonymBegin(EntitySynonymMapper):
    name = "EntitySynonymBegin"

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None
        updated_entities = message.get("entities", [])[:]
        updated_entities.sort(key=lambda x: x["start"])
        self.replace_synonyms(updated_entities)

        def shift_entities(entities, shift):
            for e in entities:
                e["start"] += shift
                e["end"] += shift

        if len(updated_entities):
            for i, entity in enumerate(updated_entities):
                literal = message.text[entity["start"] : entity["end"]]
                value = entity["value"]
                if value != literal and isinstance(value, str):
                    entity["literal"] = literal
                    message.text = (
                        message.text[0 : entity["start"]]
                        + value
                        + message.text[entity["end"] :]
                    )
                    shift = len(value) - (entity["end"] - entity["start"])
                    entity["end"] = entity["start"] + len(value)
                    if len(updated_entities) > i + 1:  # more entities:
                        shift_entities(updated_entities[i + 1 :], shift)

        message.set("entities", updated_entities, add_to_output=True)


class EntitySynonymEnd(EntitySynonymMapper):
    name = "EntitySynonymEnd"

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None
        updated_entities = message.get("entities", [])[:]
        updated_entities.sort(key=lambda x: x["start"])

        def shift_entities(entities, shift):
            for e in entities:
                e["start"] += shift
                e["end"] += shift

        for i, entity in enumerate(updated_entities):
            if "literal" in entity:
                message.text = (
                    message.text[0 : entity["start"]]
                    + entity["literal"]
                    + message.text[entity["end"] :]
                )
                shift = len(entity["literal"]) - (entity["end"] - entity["start"])
                entity["end"] = entity["start"] + len(entity["literal"])
                del entity["literal"]
                if len(updated_entities) > i + 1:  # more entities:
                    shift_entities(updated_entities[i + 1 :], shift)

        message.set("entities", updated_entities, add_to_output=True)



import logging
from typing import Text, Any, Dict, Optional, List

from rasa.core.constants import DEFAULT_REQUEST_TIMEOUT
from rasa.core.nlg.generator import NaturalLanguageGenerator
from rasa.core.trackers import DialogueStateTracker, EventVerbosity
from rasa.utils.endpoints import EndpointConfig
import os
import urllib.error

logger = logging.getLogger(__name__)


NLG_QUERY = """
fragment CarouselElementFields on CarouselElement {
    title
    subtitle
    image_url
    default_action { title, type, ...on WebUrlButton { url }, ...on PostbackButton { payload } }
    buttons { title, type, ...on WebUrlButton { url }, ...on PostbackButton { payload } }
}
query(
    $template: String!
    $arguments: Any
    $tracker: ConversationInput
    $channel: NlgRequestChannel
) {
    getResponse(
        template: $template
        arguments: $arguments
        tracker: $tracker
        channel: $channel
    ) {
        metadata
        ...on TextPayload { text }
        ...on QuickRepliesPayload { text, quick_replies { title, type, ...on WebUrlButton { url }, ...on PostbackButton { payload } } }
        ...on TextWithButtonsPayload { text, buttons { title, type, ...on WebUrlButton { url }, ...on PostbackButton { payload } } }
        ...on ImagePayload { text, image }
        ...on CarouselPayload { template_type, elements { ...CarouselElementFields } }
        ...on CustomPayload { customText: text, customImage: image, customQuickReplies: quick_replies, customButtons: buttons, customElements: elements, custom, customAttachment: attachment }
    }
}
"""


def nlg_response_format_spec():
    """Expected response schema for an NLG endpoint.

    Used for validation of the response returned from the NLG endpoint."""
    return {
        "type": "object",
        "properties": {
            "text": {"type": ["string", "null"]},
            "buttons": {"type": ["array", "null"], "items": {"type": "object"}},
            "elements": {"type": ["array", "null"], "items": {"type": "object"}},
            "attachment": {"type": ["object", "null"]},
            "image": {"type": ["string", "null"]},
        },
    }


def nlg_request_format_spec():
    """Expected request schema for requests sent to an NLG endpoint."""

    return {
        "type": "object",
        "properties": {
            "template": {"type": "string"},
            "arguments": {"type": "object"},
            "tracker": {
                "type": "object",
                "properties": {
                    "sender_id": {"type": "string"},
                    "slots": {"type": "object"},
                    "latest_message": {"type": "object"},
                    "latest_event_time": {"type": "number"},
                    "paused": {"type": "boolean"},
                    "events": {"type": "array"},
                },
            },
            "channel": {"type": "object", "properties": {"name": {"type": "string"}}},
        },
    }


def nlg_request_format(
    template_name: Text,
    tracker: DialogueStateTracker,
    output_channel: Text,
    **kwargs: Any,
) -> Dict[Text, Any]:
    """Create the json body for the NLG json body for the request."""

    tracker_state = tracker.current_state(EventVerbosity.ALL)

    return {
        "template": template_name,
        "arguments": kwargs,
        "tracker": tracker_state,
        "channel": {"name": output_channel},
    }


class GraphQLNaturalLanguageGenerator(NaturalLanguageGenerator):
    """Like Rasa's CallbackNLG, but queries Botfront's GraphQL endpoint"""

    def __init__(self, **kwargs) -> None:
        endpoint_config = kwargs.get("endpoint_config")
        self.nlg_endpoint = endpoint_config

    async def generate(
        self,
        template_name: Text,
        tracker: DialogueStateTracker,
        output_channel: Text,
        **kwargs: Any,
    ) -> List[Dict[Text, Any]]:

        fallback_language_slot = tracker.slots.get("fallback_language")
        fallback_language = (
            fallback_language_slot.initial_value if fallback_language_slot else None
        )
        language = tracker.latest_message.metadata.get("language") or fallback_language

        body = nlg_request_format(
            template_name,
            tracker,
            output_channel,
            **kwargs,
            language=language,
            projectId=os.environ.get("BF_PROJECT_ID"),
        )

        logger.debug(
            "Requesting NLG for {} from {}."
            "".format(template_name, self.nlg_endpoint.url)
        )

        try:
            if "graphql" in self.nlg_endpoint.url:
                from sgqlc.endpoint.http import HTTPEndpoint

                logging.getLogger("sgqlc.endpoint.http").setLevel(logging.WARNING)

                api_key = os.environ.get("API_KEY")
                headers = [{"Authorization": api_key}] if api_key else []
                response = HTTPEndpoint(self.nlg_endpoint.url, *headers)(
                    NLG_QUERY, body
                )
                if response.get("errors"):
                    raise urllib.error.URLError(
                        ", ".join([e.get("message") for e in response.get("errors")])
                    )
                response = response.get("data", {}).get("getResponse", {})
                if "customText" in response:
                    response["text"] = response.pop("customText")
                if "customImage" in response:
                    response["image"] = response.pop("customImage")
                if "customQuickReplies" in response:
                    response["quick_replies"] = response.pop("customQuickReplies")
                if "customButtons" in response:
                    response["buttons"] = response.pop("customButtons")
                if "customElements" in response:
                    response["elements"] = response.pop("customElements")
                if "customAttachment" in response:
                    response["attachment"] = response.pop("customAttachment")
                metadata = response.pop("metadata", {}) or {}
                for key in metadata:
                    response[key] = metadata[key]
            else:
                response = await self.nlg_endpoint.request(
                    method="post", json=body, timeout=DEFAULT_REQUEST_TIMEOUT
                )
                response = response[0]  # legacy route, use first message in seq
        except urllib.error.URLError as e:
            message = e.reason
            logger.error(f"NLG web endpoint at {self.nlg_endpoint.url} returned errors: {message}")
            return {"text": template_name}

        if self.validate_response(response):
            return response
        else:
            logger.error(f"NLG web endpoint at {self.nlg_endpoint.url} returned an invalid response.")
            return {"text": template_name}

    @staticmethod
    def validate_response(content: Optional[Dict[Text, Any]]) -> bool:
        """Validate the NLG response. Raises exception on failure."""

        from jsonschema import validate
        from jsonschema import ValidationError

        try:
            if content is None or content == "":
                # means the endpoint did not want to respond with anything
                return True
            else:
                validate(content, nlg_response_format_spec())
                return True
        except ValidationError as e:
            e.message += (
                ". Failed to validate NLG response from API, make sure your "
                "response from the NLG endpoint is valid. "
                "For more information about the format please consult the "
                "`nlg_response_format_spec` function from this same module: "
                "https://github.com/RasaHQ/rasa/blob/master/rasa/core/nlg/callback.py#L12"
            )
            raise e



import logging
import jsonpickle
import requests
import time
import os
from threading import Thread

from rasa.core.tracker_store import TrackerStore
from rasa.core.trackers import DialogueStateTracker, EventVerbosity

from sgqlc.endpoint.http import HTTPEndpoint
import urllib.error

logger = logging.getLogger(__name__)
logging.getLogger("sgqlc.endpoint.http").setLevel(logging.WARNING)

jsonpickle.set_preferred_backend("json")
jsonpickle.set_encoder_options("json", ensure_ascii=False)

GET_TRACKER = """
query trackerStore(
    $senderId: String!
    $projectId: String!
    $after: Int
    $maxEvents: Int
) {
    trackerStore(senderId: $senderId, projectId:$projectId, after:$after, maxEvents:$maxEvents) {
        tracker
        lastIndex
        lastTimestamp
    }
}
"""

INSERT_TRACKER = """
mutation insertTracker(
    $senderId: String!
    $projectId: String!
    $tracker: Any
    $env: Environement
) {
    insertTrackerStore(senderId: $senderId, projectId:$projectId, tracker:$tracker, env: $env){
        lastIndex
        lastTimestamp
    }
}
"""

UPDATE_TRACKER = """
mutation updateTracker(
    $senderId: String!
    $projectId: String!
    $tracker: Any
    $env: Environement
) {
    updateTrackerStore(senderId: $senderId, projectId: $projectId, tracker: $tracker, env: $env){
        lastIndex
        lastTimestamp
    }
}
"""


def _start_sweeper(tracker_store, break_time):
    while True:
        try:
            tracker_store.sweep()
        finally:
            time.sleep(break_time)


class BotfrontTrackerStore(TrackerStore):
    def __init__(self, domain, url, **kwargs):

        self.project_id = os.environ.get("BF_PROJECT_ID")
        self.tracker_persist_time = kwargs.get("tracker_persist_time", 3600)
        self.max_events = kwargs.get("max_events", 100)
        self.trackers = {}
        self.trackers_info = (
            {}
        )  # in this stucture we will keep the last index and the last timestamp of events in the db for a said tracker
        self.sweeper = Thread(target=_start_sweeper, args=(self, 30))
        self.sweeper.setDaemon(True)
        self.sweeper.start()
        api_key = os.environ.get("API_KEY")
        headers = [{"Authorization": api_key}] if api_key else []
        self.graphql_endpoint = HTTPEndpoint(url, *headers)
        self.url = url
        self.environement = os.environ.get("BOTFRONT_ENV", "development")

        super(BotfrontTrackerStore, self).__init__(domain)
        logger.debug("BotfrontTrackerStore tracker store created")

    def _graphql_query(self, query, params):
        try:
            response = self.graphql_endpoint(query, params)
            if response.get("errors"):
                raise urllib.error.URLError(
                    ", ".join([e.get("message") for e in response.get("errors")])
                )
            return response.get("data")
        except urllib.error.URLError as e:
            message = e.reason
            logger.error(
                f"Something went wrong getting the tracker from {self.url}: {message}"
            )
            return {}

    def _fetch_tracker(self, sender_id, lastIndex):
        data = self._graphql_query(
            GET_TRACKER,
            {
                "senderId": sender_id,
                "projectId": self.project_id,
                "after": lastIndex,
                "maxEvents": self.max_events,
            },
        )
        return data.get("trackerStore")

    def _insert_tracker_gql(self, sender_id, tracker):
        data = self._graphql_query(
            INSERT_TRACKER,
            {
                "senderId": sender_id,
                "projectId": self.project_id,
                "tracker": tracker,
                "env": self.environement,
            },
        )
        return data.get("insertTrackerStore")

    def _update_tracker_gql(self, sender_id, tracker):
        data = self._graphql_query(
            UPDATE_TRACKER,
            {
                "senderId": sender_id,
                "projectId": self.project_id,
                "tracker": tracker,
                "env": self.environement,
            },
        )
        return data.get("updateTrackerStore")

    def _get_last_index(self, sender_id):
        info = self.trackers_info.get(sender_id, -1)
        if info == -1:
            return info
        elif info.get("last_index") is None:
            return -1
        else:
            return info.get("last_index")

    def _get_last_timestamp(self, sender_id):
        info = self.trackers_info.get(sender_id, 0)
        if info == 0:
            return info
        elif info.get("last_timestamp") is None:
            return 0
        else:
            return info.get("last_timestamp")

    def _store_tracker_info(self, sender_id, tracker_info):
        if tracker_info is not None:
            self.trackers_info[sender_id] = {
                "last_index": tracker_info["lastIndex"],
                "last_timestamp": tracker_info["lastTimestamp"],
            }

    def save(self, canonical_tracker):

        # Fetch here just in case retrieve wasn't called first
        sender_id = canonical_tracker.sender_id
        tracker = self.trackers.get(sender_id)
        serialized_tracker = self._serialize_tracker_to_dict(canonical_tracker)

        if tracker is None:  # the tracker does not exist localy ( first save)
            updated_info = self._insert_tracker_gql(sender_id, serialized_tracker)
            self.trackers[sender_id] = serialized_tracker
            # update the last index and last time stamp for future uses
            self._store_tracker_info(sender_id, updated_info)
            return serialized_tracker["events"]
        else:  # the tracker  exist localy
            # Insert only the new examples
            last_timestamp = self._get_last_timestamp(sender_id)
            new_events = list(
                filter(
                    lambda x: x["timestamp"] > last_timestamp,
                    serialized_tracker["events"],
                )
            )
            tracker_shallow_copy = {key: val for key, val in serialized_tracker.items()}
            tracker_shallow_copy["events"] = new_events
            # only send the new events to the remote tracker
            updated_info = self._update_tracker_gql(sender_id, tracker_shallow_copy)
            # update the last index and last time stamp for future uses
            self._store_tracker_info(sender_id, updated_info)
            self.trackers[sender_id] = serialized_tracker
            return serialized_tracker["events"]

    def _convert_tracker(self, sender_id, tracker):
        if self.domain:
            return DialogueStateTracker.from_dict(
                sender_id, tracker["events"], self.domain.slots
            )
        else:
            logger.warning(
                "Can't recreate tracker from mongo storage "
                "because no domain is set. Returning `None` "
                "instead."
            )
            return None

    def _update_tracker(self, sender_id, remote_tracker):
        old_tracker = self.trackers.get(sender_id)
        if old_tracker is not None:
            events = old_tracker.get("events")
            remote_events = remote_tracker.get("events")
            # if we recieve max event it means that the we skiped some events
            # as we take only the last max events, so we remplace the local copy with the remote data
            if len(remote_events) == self.max_events:
                new_events = remote_events
            else:
                new_events = [*events, *remote_events]
            new_tracker = {**old_tracker, **remote_tracker}
            new_tracker["events"] = new_events
            self.trackers[sender_id] = new_tracker
            return new_tracker
        else:
            self.trackers[sender_id] = remote_tracker
            return remote_tracker

    def retrieve(self, sender_id):
        last_index = self._get_last_index(sender_id)
        # retreive all new info since the last sync (given by last index)
        new_tracker_info = self._fetch_tracker(sender_id, last_index)
        current_tracker = self.trackers.get(sender_id)
        # do not chane the order of these ifs
        # ortherwise you will get synchornication issues when working with multiple rasa instances
        # the tracker exist on the remote and may exist locally
        if new_tracker_info is not None:
            self._store_tracker_info(sender_id, new_tracker_info)
            tracker = self._update_tracker(sender_id, new_tracker_info.get("tracker"))
            return self._convert_tracker(sender_id, tracker)

        # the tracker do not exist yet
        if current_tracker is None:
            return None

        # the tracker exist localy an there is no new infos
        return self._convert_tracker(sender_id, current_tracker)

    def sweep(self):
        try:  ## wraped in a try block so if and exception occurs it does not stopp the sweep mechanism
            for key in list(
                self.trackers.keys()
            ):  # Iterate over list of keys to prevent runtime errors when deleting elements
                tracker = self.trackers.get(key)
                max_event_time = time.time() - self.tracker_persist_time
                latest_event = tracker.get("latest_event_time", float("inf"))
                if latest_event is None:
                    latest_event = float("inf")
                if tracker is not None and (latest_event < max_event_time):
                    logger.debug("SWEEPER: Removing tracker for user {}".format(key))
                    del self.trackers[key]
                    del self.trackers_info[key]
        except Exception as e:
            print(e)
            pass

    @staticmethod
    def _serialize_tracker_to_dict(canonical_tracker):
        return canonical_tracker.current_state(EventVerbosity.ALL)



import asyncio
import rasa
import logging
import inspect
from rasa.core.channels.channel import RestInput, UserMessage, CollectingOutputChannel
from sanic.request import Request
from sanic import Sanic, Blueprint, response
from asyncio import Queue, CancelledError
from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable
from rasa.core import utils
from sanic.response import HTTPResponse

logger = logging.getLogger(__name__)


class BotfrontRestOutput(CollectingOutputChannel):
    @staticmethod
    def _message(
        recipient_id: Text,
        text: Text = None,
        image: Text = None,
        buttons: List[Dict[Text, Any]] = None,
        quick_replies: List[Dict[Text, Any]] = None,
        attachment: Text = None,
        custom: Dict[Text, Any] = None,
        metadata: Dict[Text, Any] = {},
    ) -> Dict:
        """Create a message object that will be stored."""

        obj = {
            "recipient_id": recipient_id,
            "text": text,
            "image": image,
            "quick_replies": quick_replies, 
            "buttons": buttons, 
            "attachment": attachment,
            "custom": custom,
            "metadata": metadata,
        }

        # filter out any values that are `None`
        return utils.remove_none_values(obj)

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        message_parts = text.split("\n\n")
        for i in range(len(message_parts)):
            message_part = message_parts[i]
            if i == len(message_parts) - 1:
                await self._persist_message(
                    self._message(
                        recipient_id,
                        text=message_part,
                        metadata=kwargs.get("metadata", {}),
                    )
                )
            else:
                await self._persist_message(
                    self._message(recipient_id, text=message_part)
                )

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image. Default will just post the url as a string."""

        await self._persist_message(
            self._message(
                recipient_id, image=image, metadata=kwargs.get("metadata", {})
            )
        )

    async def send_attachment(
        self, recipient_id: Text, attachment: Text, **kwargs: Any
    ) -> None:
        """Sends an attachment. Default will just post as a string."""

        await self._persist_message(
            self._message(
                recipient_id, attachment=attachment, metadata=kwargs.get("metadata", {})
            )
        )

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        await self._persist_message(
            self._message(
                recipient_id,
                text=text,
                buttons=buttons,
                metadata=kwargs.get("metadata", {}),
            )
        )

    async def send_quick_replies(
        self,
        recipient_id: Text,
        text: Text,
        quick_replies: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        await self._persist_message(
            self._message(
                recipient_id,
                text=text,
                quick_replies=quick_replies,
                metadata=kwargs.get("metadata", {}),
            )
        )

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        await self._persist_message(
            self._message(
                recipient_id, custom=json_message, metadata=kwargs.get("metadata", {})
            )
        )

    async def send_elements(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        attachment = {
            "type": "template",
            "payload": {"template_type": "generic", "elements": elements},
        }
        await self._persist_message(
            self._message(
                recipient_id,
                attachment=attachment,
                metadata=kwargs.get("metadata", {}),
            )
        )


class BotfrontRestInput(RestInput):
    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        return request.json.get("customData", {})

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        custom_webhook = Blueprint(
            "custom_webhook_{}".format(type(self).__name__),
            inspect.getmodule(self).__name__,
        )

        # noinspection PyUnusedLocal
        @custom_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @custom_webhook.route("/webhook", methods=["POST"])
        async def receive(request: Request) -> HTTPResponse:
            sender_id = await self._extract_sender(request)
            text = self._extract_message(request)
            should_use_stream = rasa.utils.endpoints.bool_arg(
                request, "stream", default=False
            )
            input_channel = self._extract_input_channel(request)
            metadata = self.get_metadata(request)

            if should_use_stream:
                return response.stream(
                    self.stream_response(
                        on_new_message, text, sender_id, input_channel, metadata
                    ),
                    content_type="text/event-stream",
                )
            else:
                collector = BotfrontRestOutput()
                # noinspection PyBroadException
                try:
                    await on_new_message(
                        UserMessage(
                            text,
                            collector,
                            sender_id,
                            input_channel=input_channel,
                            metadata=metadata,
                        )
                    )
                except CancelledError:
                    logger.error(
                        "Message handling timed out for "
                        "user message '{}'.".format(text)
                    )
                except Exception:
                    logger.exception(
                        "An exception occured while handling "
                        "user message '{}'.".format(text)
                    )
                return response.json(collector.messages)

        return custom_webhook



import logging
import warnings
import uuid
from sanic import Sanic, Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse
from socketio import AsyncServer
from rasa.utils.common import raise_warning

from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable

from rasa.core.channels.channel import UserMessage, InputChannel
from rasa.core.channels.socketio import SocketIOInput, SocketIOOutput, SocketBlueprint

logger = logging.getLogger(__name__)


class WebchatOutput(SocketIOOutput):
    @classmethod
    def name(cls):
        return "webchat"

    def __init__(
        self, sio: AsyncServer, bot_message_evt: Text
    ) -> None:  # until SocketIOOutput implement comes out
        self.sio = sio
        self.bot_message_evt = bot_message_evt

    async def _send_message(self, socket_id: Text, response: Any) -> None:
        """Sends a message to the recipient using the bot event."""

        await self.sio.emit(self.bot_message_evt, response, room=socket_id)

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Send a message through this channel."""

        message_parts = text.split("\n\n")
        for i in range(len(message_parts)):
            text_message = {"text": message_parts[i]}
            if i == len(message_parts) - 1:
                text_message["metadata"] = kwargs.get("metadata", {})
            await self._send_message(recipient_id, text_message)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image to the output"""

        message = {
            "attachment": {"type": "image", "payload": {"src": image}},
            "metadata": kwargs.get("metadata", {}),
        }
        await self._send_message(recipient_id, message)

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends buttons to the output."""

        message = {
            "text": text,
            "buttons": buttons,
            "metadata": kwargs.get("metadata", {}),
        }

        await self._send_message(recipient_id, message)

    async def send_quick_replies(
        self,
        recipient_id: Text,
        text: Text,
        quick_replies: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends quick replies to the output."""

        message = {
            "text": text,
            "quick_replies": quick_replies,
            "metadata": kwargs.get("metadata", {}),
        }

        await self._send_message(recipient_id, message)

    async def send_elements(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        """Sends elements to the output."""

        message = {
            "attachment": {
                "type": "template",
                "payload": {"template_type": "generic", "elements": elements},
            },
            "metadata": kwargs.get("metadata", {}),
        }

        await self._send_message(recipient_id, message)

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends custom json to the output"""

        json_message.setdefault("room", recipient_id)

        await self.sio.emit(
            self.bot_message_evt, **json_message, metadata=kwargs.get("metadata", {})
        )

    async def send_attachment(
        self, recipient_id: Text, attachment: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends an attachment to the user."""
        await self._send_message(
            recipient_id,
            {"attachment": attachment, "metadata": kwargs.get("metadata", {})},
        )


class WebchatInput(SocketIOInput):
    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        return cls(
            credentials.get("user_message_evt", "user_uttered"),
            credentials.get("bot_message_evt", "bot_uttered"),
            credentials.get("namespace"),
            credentials.get("session_persistence", False),
            credentials.get("socketio_path", "/socket.io"),
            credentials.get("cors_allowed_origins", "*"),
        )

    @classmethod
    def name(cls):
        return "webchat"

    def __init__(
        self,
        user_message_evt: Text = "user_uttered",
        bot_message_evt: Text = "bot_uttered",
        namespace: Optional[Text] = None,
        session_persistence: bool = False,
        socketio_path: Optional[Text] = "/socket.io",
        cors_allowed_origins="*",
    ):
        self.bot_message_evt = bot_message_evt
        self.session_persistence = session_persistence
        self.user_message_evt = user_message_evt
        self.namespace = namespace
        self.socketio_path = socketio_path
        self.cors_allowed_origins = cors_allowed_origins
        self.sio = None

    def get_output_channel(self) -> Optional["OutputChannel"]:
        if self.sio is None:
            raise_warning(
                "SocketIO output channel cannot be recreated. "
                "This is expected behavior when using multiple Sanic "
                "workers or multiple Rasa Open Source instances. "
                "Please use a different channel for external events in these "
                "scenarios."
            )
            return
        return WebchatOutput(self.sio, self.bot_message_evt)

    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        return request.get("customData", {})

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        # Workaround so that socketio works with requests from other origins.
        # https://github.com/miguelgrinberg/python-socketio/issues/205#issuecomment-493769183
        sio = AsyncServer(
            async_mode="sanic", cors_allowed_origins=self.cors_allowed_origins
        )
        socketio_webhook = SocketBlueprint(
            sio, self.socketio_path, "socketio_webhook", __name__
        )

        # make sio object static to use in get_output_channel
        self.sio = sio

        @socketio_webhook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @sio.on("connect", namespace=self.namespace)
        async def connect(sid: Text, _) -> None:
            logger.debug(f"User {sid} connected to socketIO endpoint.")

        @sio.on("disconnect", namespace=self.namespace)
        async def disconnect(sid: Text) -> None:
            logger.debug(f"User {sid} disconnected from socketIO endpoint.")

        @sio.on("session_request", namespace=self.namespace)
        async def session_request(sid: Text, data: Optional[Dict]):
            if data is None:
                data = {}
            if "session_id" not in data or data["session_id"] is None:
                data["session_id"] = uuid.uuid4().hex
            if self.session_persistence:
                sio.enter_room(sid, data["session_id"])
            await sio.emit("session_confirm", data["session_id"], room=sid)
            logger.debug(f"User {sid} connected to socketIO endpoint.")

        @sio.on(self.user_message_evt, namespace=self.namespace)
        async def handle_message(sid: Text, data: Dict) -> Any:
            output_channel = WebchatOutput(sio, self.bot_message_evt)

            if self.session_persistence:
                if not data.get("session_id"):
                    warnings.warn(
                        "A message without a valid sender_id "
                        "was received. This message will be "
                        "ignored. Make sure to set a proper "
                        "session id using the "
                        "`session_request` socketIO event."
                    )
                    return
                sender_id = data["session_id"]
            else:
                sender_id = sid

            message = UserMessage(
                data["message"],
                output_channel,
                sender_id,
                input_channel=self.name(),
                metadata=self.get_metadata(data),
            )
            await on_new_message(message)

        return socketio_webhook



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from rasa.nlu.components import Component
from typing import Any, List, Optional, Text, Dict

from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.model import Metadata
from rasa.nlu.training_data import Message

logger = logging.getLogger(__name__)


class DucklingCrfMerger(Component):
    """Merges Duckling and CRF entities"""

    name = "DucklingCrfMerger"

    provides = []

    defaults = {
        "entities": None,
        "duckling_name": "DucklingHTTPExtractor",
        "extractor_names": [
            "DIETClassifier",
            "SpacyEntityExtractor",
            "CRFEntityExtractor",
            "MitieEntityExtractor",
        ],
    }

    def __init__(self, component_config=None):
        # type: (Text, Optional[List[Text]]) -> None

        super(DucklingCrfMerger, self).__init__(component_config)

    @classmethod
    def create(
        cls, component_config: Dict[Text, Any], config: RasaNLUModelConfig
    ) -> "DucklingCrfMerger":
        return cls(component_config)

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None
        crf_entities = list(
            filter(
                lambda e: e["extractor"] in self.component_config["extractor_names"]
                and e["entity"] in self.component_config["entities"].keys(),
                message.get("entities"),
            )
        )
        indices_to_remove = []

        for index, duck_entity in enumerate(message.get("entities")):
            if duck_entity["extractor"].startswith(
                self.component_config["duckling_name"]
            ):
                # looking for CRF entities surrounding the duckling one matching config settings
                containing_crf = list(
                    filter(
                        lambda e: e["start"] <= duck_entity["start"]
                        and e["end"] >= duck_entity["end"]
                        and duck_entity["entity"]
                        in self.component_config["entities"][e["entity"]],
                        crf_entities,
                    )
                )
                # list -> single object
                containing_crf = (
                    containing_crf[0]
                    if type(containing_crf) is list and len(containing_crf) > 0
                    else None
                )
                if containing_crf is not None:
                    # Add duckling value + additional infos
                    containing_crf["value"] = duck_entity["value"]
                    containing_crf["additional_info"] = duck_entity["additional_info"]
                    indices_to_remove.append(index)

        # Remove merged duckling entities
        for i in sorted(indices_to_remove, reverse=True):
            del message.get("entities")[i]

    @classmethod
    def load(
        cls,
        component_meta: Dict[Text, Any],
        model_dir: Text = None,
        model_metadata: Metadata = None,
        cached_component: Optional["DucklingCrfMerger"] = None,
        **kwargs: Any
    ) -> "DucklingCrfMerger":
        return cls(component_meta)



import logging
from typing import Text, Any, Dict, Optional, List

from rasa.core.constants import DEFAULT_REQUEST_TIMEOUT
from rasa.core.nlg.generator import NaturalLanguageGenerator
from rasa.core.trackers import DialogueStateTracker, EventVerbosity
from rasa.utils.endpoints import EndpointConfig
import os

logger = logging.getLogger(__name__)


CONFIG_QUERY = """
query(
    $projectId: String!, $environment: String
) {
    getConfig(
        projectId: $projectId,environment:$environment
    ) {
        credentials
        endpoints
    }
}
"""


async def get_config_via_graphql(bf_url, project_id):
    from sgqlc.endpoint.http import HTTPEndpoint

    logging.getLogger("sgqlc.endpoint.http").setLevel(logging.WARNING)
    import urllib.error

    environment = os.environ.get("BOTFRONT_ENV", "development")
    api_key = os.environ.get("API_KEY")
    headers = [{"Authorization": api_key}] if api_key else []
    endpoint = HTTPEndpoint(bf_url, *headers)

    async def load():
        try:
            logger.debug(f"fetching endpoints and credentials at {bf_url}")
            response = endpoint(
                CONFIG_QUERY, {"projectId": project_id, "environment": environment}
            )
            if response.get("errors"):
                raise urllib.error.URLError(
                    ", ".join([e.get("message") for e in response.get("errors")])
                )
            return endpoint(
                CONFIG_QUERY, {"projectId": project_id, "environment": environment}
            )["data"]
        except urllib.error.URLError as e:
            message = e.reason
            logger.error(f"Something went wrong fetching config at {bf_url}: {message}")
            return None

    data = await load()
    return data["getConfig"]



import json
import logging
import os
from typing import Any, List, Text, Dict
import rasa.utils.io
import re
from rasa.core.actions.action import ACTION_LISTEN_NAME
from rasa.core.constants import FALLBACK_POLICY_PRIORITY

from rasa.core.domain import Domain
from rasa.core.policies.policy import Policy, confidence_scores_for
from rasa.core.events import SlotSet
from rasa.core.trackers import DialogueStateTracker

logger = logging.getLogger(__name__)


class BotfrontDisambiguationPolicy(Policy):
    @staticmethod
    def _standard_featurizer():
        return None

    def __init__(
        self,
        priority: int = FALLBACK_POLICY_PRIORITY,
        disambiguation_trigger: str = "$0 < 2 * $1",
        fallback_trigger: float = 0.30,
        disambiguation_template: Text = "utter_disambiguation",
        excluded_intents: List = ["^chitchat\..*", "^basics\..*"],
        n_suggestions: int = 2,
    ) -> None:
        super(BotfrontDisambiguationPolicy, self).__init__(priority=priority)

        self.disambiguation_trigger = disambiguation_trigger
        self.fallback_trigger = fallback_trigger
        self.fallback_default_confidence = 0.30
        self.disambiguation_action = "action_botfront_disambiguation"
        self.disambiguation_followup_action = "action_botfront_disambiguation_followup"
        self.fallback_action = "action_botfront_fallback" # returns utter_fallback
        self.disambiguation_template = disambiguation_template
        self.excluded_intents = excluded_intents
        self.n_suggestions = n_suggestions

    def train(
        self,
        training_trackers: List[DialogueStateTracker],
        domain: Domain,
        **kwargs: Any
    ) -> None:
        pass

    def generate_disambiguation_message(self, intent_ranking, entities):
        intents = [
            (
                intent.get("name"),
                self.fill_entity(intent.get("canonical", intent.get("name")), entities)
            )
            for intent in intent_ranking
            if intent.get("name") is not None and len(
                [
                    excl
                    for excl in self.excluded_intents
                    if re.compile(excl).fullmatch(intent.get("name")) != None
                ]
            )
            < 1
        ][: self.n_suggestions]

        entities_json = (
            json.dumps({e.get("entity"): e.get("value") for e in entities})
            if len(entities) > 0
            else ""
        )

        quick_replies = []
        for intent in intents:
            quick_replies.append(
                {
                    "title": intent[1],
                    "type": "postback",
                    "payload": "/{}{}".format(intent[0], entities_json),
                }
            )
        return {"template": self.disambiguation_template, "quick_replies": quick_replies}

    @staticmethod
    def fill_entity(template, entities):
        title = template
        for e in entities:
            title = title.replace("{" + e.get("entity") + "}", e.get("value"))
        title = re.sub(r"{.*?}", r"", title)
        return title

    @staticmethod
    def set_slot(tracker, message):
        if len(message["quick_replies"]) < 2:
            return None  # abort if only deny_suggestions button would be shown
        try:
            tracker.update(SlotSet("disambiguation_message", value=message))
            result = message
        except Exception as e:
            logger.error("Could not set message slot: {}".format(e))
            result = None
        return result

    @staticmethod
    def _should_disambiguate(intent_ranking, trigger):
        import re

        # pattern to match $0, $1, $2, ... and returning 0, 1, 2,... in match groups
        pattern = re.compile(r"\$(\d)")
        eval_string = trigger
        # matches: an array of intents indices to consider in intent_ranking
        matches = re.findall(pattern, trigger)
        for i in matches:
            # if not enough intents in ranking to apply the rule, policy rule can't be triggered
            if int(i) >= len(intent_ranking):
                return False
            eval_string = re.sub(
                r"\$" + i, str(intent_ranking[int(i)]["confidence"]), eval_string
            )

        return eval(eval_string, {"__builtins__": {}})

    @staticmethod
    def _should_fallback(intent_ranking, trigger):
        bonified_ranking = intent_ranking + [{"confidence": 0}]
        return bonified_ranking[0]["confidence"] < trigger

    def _is_user_input_expected(self, tracker: DialogueStateTracker) -> bool:
        return tracker.latest_action_name in [
            self.fallback_action,
            self.disambiguation_action,
        ]

    def _have_options_been_suggested(self, tracker: DialogueStateTracker) -> bool:
        return tracker.last_executed_action_has(self.disambiguation_action)

    def predict_action_probabilities(
        self, tracker: DialogueStateTracker, domain: Domain
    ) -> List[float]:

        parse_data = tracker.latest_message.parse_data
        entities = parse_data.get("entities", [])
        intent_ranking = parse_data.get("intent_ranking", [])
        if len(intent_ranking) == 0 and parse_data.get("intent") is not None:
            intent_ranking = [parse_data.get("intent")]
        can_apply = tracker.latest_action_name == ACTION_LISTEN_NAME
        should_fallback = can_apply and self._should_fallback(
            intent_ranking, self.fallback_trigger
        )
        should_disambiguate = can_apply and self._should_disambiguate(
            intent_ranking, self.disambiguation_trigger
        )

        if self._is_user_input_expected(tracker):
            # Shut up and listen
            result = confidence_scores_for(ACTION_LISTEN_NAME, 1.0, domain)

        elif should_fallback:
            logger.debug("Triggering fallback")
            result = confidence_scores_for(self.fallback_action, 1.0, domain)

        elif self._have_options_been_suggested(tracker):
            if not should_disambiguate:
                logger.debug("Successfully disambiguated")
                result = confidence_scores_for(
                    self.disambiguation_followup_action, 1.0, domain
                )
            else:
                logger.debug(
                    "Will not disambiguate a second time so fast -- triggering fallback"
                )
                result = confidence_scores_for(self.fallback_action, 1.0, domain)

        elif should_disambiguate:
            logger.debug("Triggering disambiguation")
            disambiguation_message = self.generate_disambiguation_message(
                intent_ranking, entities,
            )
            slot_set = self.set_slot(tracker, disambiguation_message)
            if slot_set:
                result = confidence_scores_for(self.disambiguation_action, 1.0, domain)
            else:
                result = confidence_scores_for(self.fallback_action, 1.0, domain)

        else:
            # Nothing to see here; setting fallback to default confidence
            result = confidence_scores_for(
                self.fallback_action, self.fallback_default_confidence, domain
            )

        return result

    def persist(self, path: Text) -> None:
        """Persists the policy to storage."""

        config_file = os.path.join(path, "botfront_disambiguation_policy.json")
        meta = {
            "priority": self.priority,
            "disambiguation_trigger": self.disambiguation_trigger,
            "fallback_trigger": self.fallback_trigger,
            "disambiguation_template": self.disambiguation_template,
            "excluded_intents": self.excluded_intents,
            "n_suggestions": self.n_suggestions,
        }
        rasa.utils.io.create_directory_for_file(config_file)
        rasa.utils.io.dump_obj_as_json_to_file(config_file, meta)

    @classmethod
    def load(cls, path: Text) -> "BotfrontDisambiguationPolicy":
        meta = {}
        if os.path.exists(path):
            meta_path = os.path.join(path, "botfront_disambiguation_policy.json")
            if os.path.isfile(meta_path):
                meta = json.loads(rasa.utils.io.read_file(meta_path))

        return cls(**meta)



import asyncio
import rasa
import logging
import inspect
import os
from rasa.core.channels.channel import (
    RestInput,
    UserMessage,
    CollectingOutputChannel,
    InputChannel,
)
from sanic.request import Request
from sanic import Sanic, Blueprint, response
from asyncio import Queue, CancelledError
from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable
from rasa.core import utils
from sanic.response import HTTPResponse
from rasa_addons.core.channels.rest import BotfrontRestInput, BotfrontRestOutput
from rasa.utils.endpoints import EndpointConfig
from rasa_addons.core.channels.graphql import get_config_via_graphql

logger = logging.getLogger(__name__)


class BotfrontRestPlusInput(BotfrontRestInput):
    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        credentials = credentials or {}
        return cls(credentials.get("config"))

    def __init__(self, config: Optional[Dict[Text, Any]] = None):
        self.config = config

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        custom_webhook = Blueprint(
            "custom_webhook_{}".format(type(self).__name__),
            inspect.getmodule(self).__name__,
        )

        # noinspection PyUnusedLocal
        @custom_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @custom_webhook.route("/props", methods=["GET"])
        async def serve_rules(request: Request) -> HTTPResponse:
            if self.config:
                return response.json(self.config)
            else:
                config = await get_config_via_graphql(
                    os.environ["BF_URL"], os.environ["BF_PROJECT_ID"]
                )
                return response.json(
                    config["credentials"][
                        "rasa_addons.core.channels.rest_plus.BotfrontRestPlusInput"
                    ]["props"]
                )


        @custom_webhook.route("/webhook", methods=["POST"])
        async def receive(request: Request) -> HTTPResponse:
            sender_id = await self._extract_sender(request)
            text = self._extract_message(request)
            should_use_stream = rasa.utils.endpoints.bool_arg(
                request, "stream", default=False
            )
            input_channel = self._extract_input_channel(request)
            metadata = self.get_metadata(request)

            if should_use_stream:
                return response.stream(
                    self.stream_response(
                        on_new_message, text, sender_id, input_channel, metadata
                    ),
                    content_type="text/event-stream",
                )
            else:
                collector = BotfrontRestOutput()
                # noinspection PyBroadException
                try:
                    await on_new_message(
                        UserMessage(
                            text,
                            collector,
                            sender_id,
                            input_channel=input_channel,
                            metadata=metadata,
                        )
                    )
                except CancelledError:
                    logger.error(
                        "Message handling timed out for "
                        "user message '{}'.".format(text)
                    )
                except Exception:
                    logger.exception(
                        "An exception occured while handling "
                        "user message '{}'.".format(text)
                    )
                return response.json(collector.messages)

        return custom_webhook



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os

import requests
from rasa.nlu.components import Component
from typing import Any, List, Optional, Text, Dict

from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.model import Metadata
from rasa.nlu.training_data import Message

logger = logging.getLogger(__name__)


class EntitiesFilter(Component):
    """Filter entities wrt intent"""

    name = "EntitiesFilter"
    defaults = {
        "entities": {},
        "extractor_names": [
            "DIETClassifier",
            "SpacyEntityExtractor",
            "CRFEntityExtractor",
            "MitieEntityExtractor",
        ],
    }

    def __init__(self, component_config=None):
        # type: (Text, Optional[List[Text]]) -> None

        super(EntitiesFilter, self).__init__(component_config)

    @classmethod
    def create(
        cls, component_config: Dict[Text, Any], config: RasaNLUModelConfig
    ) -> "EntitiesFilter":
        return cls(component_config)

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        # get intent
        intent = message.get("intent")
        if intent is None:
            logger.warn("No intent found")

        # get crf and duckling entities
        message_entities = message.get("entities")
        entities_to_filter = filter(
            lambda e: e["extractor"] in self.component_config["extractor_names"],
            message_entities,
        )
        indices_to_remove = []
        for index, entity in enumerate(entities_to_filter):
            if (
                intent["name"] in self.component_config["entities"].keys()
                and entity["entity"]
                not in self.component_config["entities"][intent["name"]]
            ):
                indices_to_remove.append(index)

        for i in sorted(indices_to_remove, reverse=True):
            del message.get("entities")[i]

    @classmethod
    def load(
        cls,
        component_meta: Dict[Text, Any],
        model_dir: Text = None,
        model_metadata: Metadata = None,
        cached_component: Optional["EntitiesFilter"] = None,
        **kwargs: Any
    ) -> "EntitiesFilter":
        return cls(component_meta)



import os
import warnings

from typing import Any, Text, Dict, Optional, List

from rasa.nlu.components import Component
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.training_data import Message, TrainingData
from rasa.nlu.model import Metadata


class IntentRankingCanonicalExampleInjector(Component):
    """
        Will remember the first encountered intent/entity-value combination
        in the training data, and inject it under a "canonical" key in each
        member of intent_ranking.
    """

    name = "IntentRankingCanonicalExampleInjector"

    defaults = {}

    def __init__(
        self,
        component_config: Optional[Dict[Text, Any]] = None,
        canonicals: Optional[Dict] = {},
    ) -> None:

        super(IntentRankingCanonicalExampleInjector, self).__init__(component_config)
        self.canonicals = canonicals

    @staticmethod
    def generate_entity_pairs(entities):
        return frozenset((e.get("entity"), e.get("value")) for e in entities)

    def generate_canonicals(self, nlu_data):
        canonicals = {}
        for datum in nlu_data:
            intent, text, entities = (
                datum.get("intent", ""),
                datum.get("text", ""),
                datum.get("entities", []),
            )
            entities = self.generate_entity_pairs(entities)
            if intent not in canonicals:
                canonicals[intent] = {}
            if entities not in canonicals[intent]:
                canonicals[intent][entities] = text

        return canonicals

    def train(
        self, training_data: TrainingData, cfg: RasaNLUModelConfig, **kwargs: Any
    ) -> None:

        self.canonicals = self.generate_canonicals(training_data.training_examples)

    def get_canonical(self, intent, entities):
        if intent not in self.canonicals.keys():
            return None
        entities = self.generate_entity_pairs(entities)
        canonicals = self.canonicals[intent]
        if len(canonicals or {}) < 1:
            return None
        if entities in canonicals:
            return canonicals[entities]
        matches = sorted(
            canonicals.keys(), key=lambda k: len(k.symmetric_difference(entities)),
        )
        return canonicals[matches[0]]

    def process(self, message: Message, **kwargs: Any) -> None:
        intent_ranking, entities = (
            message.get("intent_ranking", []),
            message.get("entities", []),
        )

        for i in range(len(intent_ranking)):
            intent_ranking[i]["canonical"] = self.get_canonical(
                intent_ranking[i].get("name"), entities
            )

        message.set(
            "intent_ranking", intent_ranking, add_to_output=True,
        )

    @classmethod
    def load(
        cls,
        meta: Dict[Text, Any],
        model_dir: Text = None,
        model_metadata: Metadata = None,
        cached_component: Optional["IntentRankingCanonicalExampleInjector"] = None,
        **kwargs: Any
    ) -> "IntentRankingCanonicalExampleInjector":

        from rasa.utils.io import pickle_load

        file_name = meta.get("file")
        path = os.path.join(model_dir, file_name)

        if os.path.exists(path):
            return cls(meta, **pickle_load(path))
        else:
            return cls(meta)

    def persist(self, file_name: Text, model_dir: Text) -> Optional[Dict[Text, Any]]:

        from rasa.utils.io import pickle_dump

        file_name = file_name + ".pickle"
        path = os.path.join(model_dir, file_name)
        persisted = {"canonicals": self.canonicals}
        pickle_dump(path, persisted)

        return {"file": file_name}



import logging
import warnings
import uuid
from sanic import Sanic, Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse
from socketio import AsyncServer
import os


from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable

from rasa.core.channels.channel import UserMessage
from rasa.core.channels.socketio import SocketIOInput, SocketIOOutput, SocketBlueprint

from rasa.core.channels.channel import InputChannel
from rasa_addons.core.channels.webchat import WebchatOutput
from rasa_addons.core.channels.webchat import WebchatInput
from rasa_addons.core.channels.graphql import get_config_via_graphql

logger = logging.getLogger(__name__)


class WebchatPlusInput(WebchatInput):
    @classmethod
    def name(cls):
        return "webchat"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        credentials = credentials or {}
        return cls(
            credentials.get("user_message_evt", "user_uttered"),
            credentials.get("bot_message_evt", "bot_uttered"),
            credentials.get("namespace"),
            credentials.get("session_persistence", False),
            credentials.get("socketio_path", "/socket.io"),
            credentials.get("cors_allowed_origins", "*"),
            credentials.get("config"),
        )

    def __init__(
        self,
        user_message_evt: Text = "user_uttered",
        bot_message_evt: Text = "bot_uttered",
        namespace: Optional[Text] = None,
        session_persistence: bool = False,
        socketio_path: Optional[Text] = "/socket.io",
        cors_allowed_origins="*",
        config=None,
    ):
        self.bot_message_evt = bot_message_evt
        self.session_persistence = session_persistence
        self.user_message_evt = user_message_evt
        self.namespace = namespace
        self.socketio_path = socketio_path
        self.cors_allowed_origins = cors_allowed_origins
        self.config = config

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        # Workaround so that socketio works with requests from other origins.
        # https://github.com/miguelgrinberg/python-socketio/issues/205#issuecomment-493769183
        sio = AsyncServer(
            async_mode="sanic", cors_allowed_origins=self.cors_allowed_origins
        )
        socketio_webhook = SocketBlueprint(
            sio, self.socketio_path, "socketio_webhook", __name__
        )

        @socketio_webhook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @sio.on("connect", namespace=self.namespace)
        async def connect(sid: Text, _) -> None:
            logger.debug(f"User {sid} connected to socketIO endpoint.")

        @sio.on("disconnect", namespace=self.namespace)
        async def disconnect(sid: Text) -> None:
            logger.debug(f"User {sid} disconnected from socketIO endpoint.")

        @sio.on("session_request", namespace=self.namespace)
        async def session_request(sid: Text, data: Optional[Dict]):
            if data is None:
                data = {}
            if "session_id" not in data or data["session_id"] is None:
                data["session_id"] = uuid.uuid4().hex

            if self.config is not None:
                props = self.config
            else:
                config = await get_config_via_graphql(
                    os.environ["BF_URL"], os.environ["BF_PROJECT_ID"]
                )
                props = config["credentials"][
                    "rasa_addons.core.channels.webchat_plus.WebchatPlusInput"
                ]["props"]

            await sio.emit(
                "session_confirm",
                {"session_id": data["session_id"], "props": props},
                room=sid,
            )
            logger.debug(f"User {sid} connected to socketIO endpoint.")

        @sio.on(self.user_message_evt, namespace=self.namespace)
        async def handle_message(sid: Text, data: Dict) -> Any:
            output_channel = WebchatOutput(sio, sid, self.bot_message_evt)

            if self.session_persistence:
                if not data.get("session_id"):
                    warnings.warn(
                        "A message without a valid sender_id "
                        "was received. This message will be "
                        "ignored. Make sure to set a proper "
                        "session id using the "
                        "`session_request` socketIO event."
                    )
                    return
                sender_id = data["session_id"]
            else:
                sender_id = sid

            message = UserMessage(
                data["message"],
                output_channel,
                sender_id,
                input_channel=self.name(),
                metadata=self.get_metadata(data),
            )
            await on_new_message(message)

        return socketio_webhook




import logging
import os
import copy
from typing import Optional, Text, Union, List, Dict

from rasa import data
from rasa.core.domain import Domain, InvalidDomain
from rasa.core.interpreter import RegexInterpreter, NaturalLanguageInterpreter
from rasa.core.training.structures import StoryGraph
from rasa.core.training.dsl import StoryFileReader
from rasa.importers import utils
from rasa.importers.importer import TrainingDataImporter
from rasa.nlu.training_data import TrainingData
from rasa.utils import io as io_utils
from rasa.core.utils import get_file_hash

logger = logging.getLogger(__name__)


class BotfrontFileImporter(TrainingDataImporter):
    def __init__(
        self,
        config_file: Optional[Union[List[Text], Text]] = None,
        domain_path: Optional[Text] = None,
        training_data_paths: Optional[Union[List[Text], Text]] = None,
    ):
        self._domain_path = domain_path

        self._story_files, self._nlu_files = data.get_core_nlu_files(
            training_data_paths
        )

        self.core_config = {}
        self.nlu_config = {}
        if config_file:
            if not isinstance(config_file, list): config_file = [config_file]
            for file in config_file:
                if not os.path.exists(file): continue
                config = io_utils.read_config_file(file)
                lang = config["language"]
                self.core_config = {"policies": config["policies"]}
                self.nlu_config[lang] = {"pipeline": config["pipeline"], "data": lang}
    
    def path_for_nlu_lang(self, lang) -> Text:
        return [x for x in self._nlu_files if "{}.md".format(lang) in x]

    async def get_config(self) -> Dict:
        return self.core_config

    async def get_nlu_config(self, languages=True) -> Dict:
        if not isinstance(languages, list):
            languages = self.nlu_config.keys()
        return {
            lang: self.nlu_config[lang] if lang in languages else False
            for lang in self.nlu_config.keys()
        }

    async def get_stories(
        self,
        interpreter: "NaturalLanguageInterpreter" = RegexInterpreter(),
        template_variables: Optional[Dict] = None,
        use_e2e: bool = False,
        exclusion_percentage: Optional[int] = None,
    ) -> StoryGraph:

        story_steps = await StoryFileReader.read_from_files(
            self._story_files,
            await self.get_domain(),
            interpreter,
            template_variables,
            use_e2e,
            exclusion_percentage,
        )
        return StoryGraph(story_steps)

    async def get_nlu_data(self, languages=True) -> Dict[Text, TrainingData]:
        language = None
        if isinstance(languages, str):
            language = languages
            languages = [language]
        if not isinstance(languages, list):
            languages = self.nlu_config.keys()
        td = {}
        for lang in languages:
            try:
                td[lang] = utils.training_data_from_paths(
                    self.path_for_nlu_lang(lang), lang,
                )
            except ValueError as e:
                if str(e).startswith("Unknown data format"):
                    td[lang] = TrainingData()
        if language: return td.get(language, TrainingData())
        return td

    async def get_domain(self) -> Domain:
        domain = Domain.empty()
        try:
            domain = Domain.load(self._domain_path)
            domain.check_missing_templates()
        except InvalidDomain as e:
            logger.warning(
                "Loading domain from '{}' failed. Using empty domain. Error: '{}'".format(
                    self._domain_path, e.message
                )
            )

        return domain



import logging

logger = logging.getLogger(__name__)


def get_latest_parse_data_language(all_events):
    events = reversed(all_events)
    try:
        while True:
            event = next(events)
            if (
                event["event"] == "user"
                and "parse_data" in event
                and "language" in event["parse_data"]
            ):
                return event["parse_data"]["language"]

    except StopIteration:
        return None



from setuptools import setup, find_packages
import os
import io

with io.open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"),
    encoding="utf-8",
) as f:
    long_description = f.read()
setup(
    name="rasa_addons",
    version="0.9.29.2",
    author="Botfront",
    description="Rasa Addons - Components for Rasa and Botfront",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "requests",
        "requests_futures",
        "fuzzy_matcher",
        "fbmessenger",
        "sgqlc",
    ],
    packages=find_packages(exclude=["tests"]),
    licence="Apache 2.0",
    url="https://botfront.io",
    author_email="hi@botfront.io",
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests
import logging
import json

from typing import Any

from rasa.nlu.components import Component
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.training_data.message import Message

logger = logging.getLogger(__name__)


class Sweeper(Component):
    name = "Sweeper"

    defaults = {"entity_names": []}

    def __init__(self, component_config=None):
        # type: (RasaNLUModelConfig) -> None
        super(Sweeper, self).__init__(component_config)

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        message_entities = message.get("entities")
        sweeped_entities = list(
            filter(
                lambda e: e["entity"] not in self.component_config["entity_names"],
                message_entities,
            )
        )
        message.set("entities", sweeped_entities)

